#include "base_controller/BaseController.hpp"
#include <tf2/LinearMath/Quaternion.h>
#include <tf2/LinearMath/Matrix3x3.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <Eigen/QR>

BaseController::BaseController(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    if(!node_handle_.getParam("/husky_velocity_controller/linear/x/max_velocity", max_linear_velocity_))
    {
        ROS_ERROR("Could not retrieve parameter /husky_velocity_controller/linear/x/max_velocity");
    }
    if(!node_handle_.getParam("/husky_velocity_controller/angular/z/max_velocity", max_angular_velocity_))
    {
        ROS_ERROR("Could not retrieve parameter /husky_velocity_controller/angular/z/max_velocity");
    }
    husky_vel_publisher_ = node_handle_.advertise<geometry_msgs::Twist>("/husky_velocity_controller/cmd_vel", 1, this);
    target_twist_subscriber_ = node_handle_.subscribe("/planner/twist", 1, &BaseController::target_twist_callback, this);
    target_pose_subscriber_ = node_handle_.subscribe("/planner/pose", 1, &BaseController::target_pose_callback, this);
    feedback_pose_subscriber_ = node_handle_.subscribe("/husky_velocity_controller/feedback/pose", 1, &BaseController::feedback_pose_callback, this);
    planning_status_subscriber_ = node_handle_.subscribe("/planner/done", 1, &BaseController::planning_status_callback, this);
    is_planning_done_ = true; // Planning not started yet
}

void BaseController::target_pose_callback(const geometry_msgs::Pose& msg)
{
    target_position_(0) = msg.position.x;
    target_position_(1) = msg.position.y;
}

void BaseController::target_twist_callback(const geometry_msgs::Twist& msg)
{
    target_velocity_(0) = msg.linear.x;
    target_velocity_(1) = msg.linear.y;
}

void BaseController::feedback_pose_callback(const geometry_msgs::Pose& msg)
{
    current_position_(0) = msg.position.x;
    current_position_(1) = msg.position.y;
    tf2::Quaternion tf_orientation;
    tf2::fromMsg(msg.orientation, tf_orientation);
    double roll, pitch, yaw;
    tf2::Matrix3x3(tf_orientation).getRPY(roll, pitch, yaw);
    current_heading_ = yaw;
}

void BaseController::planning_status_callback(const std_msgs::Bool& msg)
{
    if (!is_planning_done_ && msg.data) // Check for planning status change from not done to done
    {
        ROS_INFO_STREAM("Reached target x: " << target_position_(0) << ", y: " << target_position_(1) << ". Stopping the robot.");
    }
    is_planning_done_ = msg.data;
}

Eigen::MatrixXd BaseController::compute_jacobian(double x_r, double y_r, double phi)
{
    Eigen::MatrixXd jacobian = Eigen::MatrixXd::Zero(2, 2);
    jacobian(0, 0) = cos(phi);
    jacobian(0, 1) = -sin(phi) * x_r - cos(phi) * y_r;
    jacobian(1, 0) = sin(phi);
    jacobian(1, 1) = cos(phi) * x_r - sin(phi) * y_r;
    return jacobian;
}

void BaseController::compute_twist()
{
    Eigen::MatrixXd jacobian = compute_jacobian(target_velocity_(0), target_velocity_(1), current_heading_);
    Eigen::Vector2d twist = jacobian.completeOrthogonalDecomposition().pseudoInverse() * target_velocity_;
    if(abs(twist(0)) > max_linear_velocity_)
    {
        twist(0) = twist(0) / abs(twist(0)) * max_linear_velocity_;
    }
    if(abs(twist(1)) > max_angular_velocity_)
    {
        twist(1) = twist(1) / abs(twist(1)) * max_angular_velocity_;
    }
    husky_vel_msg_.linear.x = twist(0);
    husky_vel_msg_.angular.z = twist(1);
}

void BaseController::reset_twist()
{
    husky_vel_msg_.linear.x = 0;
    husky_vel_msg_.angular.z = 0;
}

void BaseController::update()
{
    if(!is_planning_done_)
    {
        compute_twist();
    } else 
    {
        reset_twist();
    }
    husky_vel_publisher_.publish(husky_vel_msg_);
}