#include "cubic_polynomial_planner/CubicActionServer.hpp"

CubicActionServer::CubicActionServer(ros::NodeHandle& node_handle) : node_handle_(node_handle), action_server_(node_handle_, "pose_command", boost::bind(&CubicActionServer::cubic_action_callback, this, _1), false)
{
    if (!load_params())
    {
        ROS_ERROR("Could not load parameters");
        ros::requestShutdown();
    }
    // MIGHT NEED TO REMAP 
    pose_ref_publisher_ = node_handle_.advertise<geometry_msgs::Pose>("pose_cmd", 1, this);
    twist_ref_publisher_ = node_handle_.advertise<geometry_msgs::Twist>("twist_cmd", 1, this);
    pose_fbk_subscriber_ = node_handle_.subscribe(pose_subscriber_topic_, 1, &CubicActionServer::pose_fbk_callback, this);
    // REMAP THIS TO THE ACTION SERVER
    // action_server_ = actionlib::SimpleActionServer<highlevel_msgs::PoseCommandAction>(node_handle_, "pose_command", boost::bind(&CubicActionServer::cubic_action_callback, this, _1), false);

    ground_truth_position_ = Eigen::VectorXd::Zero(3);
    initial_position_ = Eigen::VectorXd::Zero(3);
    target_position_ = Eigen::VectorXd::Zero(3);
    next_position_ = Eigen::VectorXd::Zero(3);
    next_velocity_ = Eigen::VectorXd::Zero(3);
    target_ori_ = Eigen::Quaterniond::Identity();
    initial_ori_ = Eigen::Quaterniond::Identity();
    ground_truth_ori_ = Eigen::Quaterniond::Identity();
    next_ori_ = Eigen::Quaterniond::Identity();
    target_T_ = -1; // Ensure current time is greater than target_T_ initially
    is_pose_available_ = false;
    is_initialised_ = false;
    action_server_.start();
    ROS_INFO("Cubic action server started");
}

bool CubicActionServer::load_params()
{
    if (!node_handle_.getParam("pose_subscriber_topic", pose_subscriber_topic_))
    {
        ROS_ERROR("Could not load parameter pose_subscriber_topic");
        return false;
    }
    if (!node_handle_.getParam("z_min", z_min_))
    {
        ROS_ERROR("Could not load parameter z_min");
        return false;
    }
    if (!node_handle_.getParam("/publish_rate", loop_rate_))
    {
        ROS_ERROR("Could not load parameter /publish_rate");
        return false;
    }
    return true;
}

void CubicActionServer::pose_fbk_callback(const geometry_msgs::Pose& pose_msg)
{
    ground_truth_position_ << pose_msg.position.x, pose_msg.position.y, pose_msg.position.z;
    ground_truth_ori_ = Eigen::Quaterniond(pose_msg.orientation.w, pose_msg.orientation.x, pose_msg.orientation.y, pose_msg.orientation.z);
    // fbk_quat_ = tf2::Quaternion(pose_msg.orientation.x, pose_msg.orientation.y, pose_msg.orientation.z, pose_msg.orientation.w);
    tf2::fromMsg(pose_msg.orientation, fbk_quat_);
    is_pose_available_ = true;
}

void CubicActionServer::compute_trajectory_point()
{
    double current_T = ros::Time::now().toSec();
    if (current_T <= target_T_)
    {
        double time_elapsed = current_T - initial_T_;
        s_ = a0_ + a1_*time_elapsed + a2_*pow(time_elapsed, 2) + a3_*pow(time_elapsed, 3);
        s_dot_ = a1_ + 2*a2_*time_elapsed + 3*a3_*pow(time_elapsed, 2);
        next_position_ = initial_position_ + s_ * (target_position_ - initial_position_);
        next_velocity_ = s_dot_ * (target_position_ - initial_position_);
        next_ori_ = initial_ori_.slerp(s_, target_ori_); 

        pose_cmd_.position.x = next_position_(0);
        pose_cmd_.position.y = next_position_(1);
        pose_cmd_.position.z = next_position_(2);
        pose_cmd_.orientation.x = next_ori_.x();
        pose_cmd_.orientation.y = next_ori_.y();
        pose_cmd_.orientation.z = next_ori_.z();
        pose_cmd_.orientation.w = next_ori_.w();
        
        twist_cmd_.linear.x = next_velocity_(0);
        twist_cmd_.linear.y = next_velocity_(1);
        twist_cmd_.linear.z = next_velocity_(2);
    }
}

void CubicActionServer::cubic_action_callback(const highlevel_msgs::PoseCommandGoalConstPtr& goal)
{
    ROS_INFO_STREAM("Cubic action server called: Move to x: " << goal->x << ", y: " << goal->y << ", z: " << goal->z << ", T: " << goal->T);
    ROS_INFO_STREAM("From x: " << ground_truth_position_(0) << ", y: " << ground_truth_position_(1) << ", z: " << ground_truth_position_(2));

    ros::Rate loop_rate(loop_rate_);

    while (!is_pose_available_)
    {
        ROS_ERROR("Pose not available, try again later");
        loop_rate.sleep();
    }

    ROS_INFO_STREAM(ground_truth_position_.transpose());
    // Set initial values 
    initial_position_ = ground_truth_position_;
    initial_ori_ = ground_truth_ori_;
    initial_T_ = ros::Time::now().toSec();

    target_position_ << goal->x, goal->y, goal->z;
    target_quat_.setRPY(goal->roll, goal->pitch, goal->yaw);
    target_ori_ = Eigen::Quaterniond(target_quat_.w(), target_quat_.x(), target_quat_.y(), target_quat_.z());
    target_T_ = initial_T_ + goal->T;
    
    // Cubic polynomial coefficients
    a0_ = 0; // Initial displacement zero
    a1_ = 0; // Initial velocity zero
    a2_ = 3 / pow(goal->T, 2);
    a3_ = -2 / pow(goal->T, 3);

    ROS_INFO("Beginning trajectory generation");

    double time_elapsed = 0;
    while (time_elapsed <= goal->T)
    {
        compute_trajectory_point();
        pose_ref_publisher_.publish(pose_cmd_);
        twist_ref_publisher_.publish(twist_cmd_);
        action_server_.publishFeedback(action_feedback_);
        loop_rate.sleep();
        time_elapsed = ros::Time::now().toSec() - initial_T_;
    }

    ROS_INFO("Trajectory generation complete");
    action_server_.setSucceeded();
}

void CubicActionServer::update()
{
    double time_elapsed = ros::Time::now().toSec() - initial_T_;
    double distance_translation = (target_position_ - ground_truth_position_).norm();
    double distance_orientation = tf2::angleShortestPath(fbk_quat_, target_quat_);
    action_feedback_.time_elapsed = time_elapsed;
    action_feedback_.distance_translation = distance_translation;
    action_feedback_.distance_orientation = distance_orientation;
}