#include "cubic_polynomial_planner/CubicPolynomialPlanner.hpp"

CubicPolynomialPlanner::CubicPolynomialPlanner(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    ROS_INFO("Preparing pose_planner/move_to service.");
    // Get parameters from parameter server
    if(!load_params())
    {
        ROS_ERROR("Could  not load parameters");
        ros::requestShutdown();
    }
    move_to_server_ = node_handle_.advertiseService("move_to", &CubicPolynomialPlanner::move_to_callback, this);
    move_ori_server_ = node_handle_.advertiseService("move_ori", &CubicPolynomialPlanner::move_ori_callback, this);
    pose_subscriber_ = node_handle_.subscribe(pose_subscriber_topic_, 1, &CubicPolynomialPlanner::pose_callback, this);
    pose_publisher_ = node_handle_.advertise<geometry_msgs::Pose>("pose_cmd", 1, this);
    twist_publisher_ = node_handle_.advertise<geometry_msgs::Twist>("twist_cmd", 1, this);
    planning_status_publisher_ = node_handle_.advertise<std_msgs::Bool>("done", 1, this);
    ground_truth_position_ = Eigen::VectorXd::Zero(3);
    initial_position_ = Eigen::VectorXd::Zero(3);
    target_position_ = Eigen::VectorXd::Zero(3);
    next_position_ = Eigen::VectorXd::Zero(3);
    next_velocity_ = Eigen::VectorXd::Zero(3);
    target_ori_ = Eigen::Quaterniond::Identity();
    initial_ori_ = Eigen::Quaterniond::Identity();
    ground_truth_ori_ = Eigen::Quaterniond::Identity();
    next_ori_ = Eigen::Quaterniond::Identity();
    target_T_ = -1; // Ensure current time is greater than target_T_ initially
    is_pose_available_ = false;
    is_initialised_ = false;
    planning_status_msg_.data = true; // Free for next service call
}

bool CubicPolynomialPlanner::load_params()
{
    if(!node_handle_.getParam("pose_subscriber_topic", pose_subscriber_topic_))
    {
        ROS_ERROR("Could not load parameter pose_subscriber_topic");
        return false;
    }
    if(!node_handle_.getParam("z_min", z_min_))
    {
        ROS_ERROR("Could not load parameter z_min");
        return false;
    }
    return true;
}

bool CubicPolynomialPlanner::move_to_callback(highlevel_msgs::MoveTo::Request& req, highlevel_msgs::MoveTo::Response& resp)
{
    ROS_INFO("Planner service called: Move to x: %f, y: %f, z: %f, T: %f.", req.x, req.y, req.z, req.T);
    ROS_INFO("From x: %f, y: %f, z: %f, T: %f.", ground_truth_position_(0), ground_truth_position_(1), ground_truth_position_(2));

    if (!is_pose_available_)
    {
        ROS_ERROR("Pose not available, try again later");
        resp.success = false;
        return false;
    }

    // Input checking
    if (req.T <= 0)
    {
        ROS_ERROR("Time for planning must be greater than 0, but entered time: %f.", req.T);
        resp.success = false;
        return false;
    }
    if (req.z < z_min_)
    {
        ROS_ERROR("Target z must be greater than 0.5, but entered z: %f.", req.z);
        resp.success = false;
        return false;
    }

    // Set initial values upon service call
    initial_position_ = ground_truth_position_;
    initial_T_ = ros::Time::now().toSec();

    // Extract target values
    target_position_ << req.x, req.y, req.z;
    target_T_ = initial_T_ + req.T;

    if (!is_initialised_)
    {
        initial_ori_ = ground_truth_ori_;
        target_ori_ = initial_ori_;
        is_initialised_ = true;
    } else 
    {
        initial_ori_ = target_ori_; // Keep last target orientation
    }

    // Compute polynomial coefficients
    a0_ = 0; // Initial displacement zero
    a1_ = 0; // Initial velocity zero
    a2_ = 3 / pow(req.T, 2);
    a3_ = -2 / pow(req.T, 3);

    ROS_INFO("Request valid, trajectory execution will begin shortly.");
    resp.success = true;
    return true;
}

bool CubicPolynomialPlanner::move_ori_callback(highlevel_msgs::MoveTo::Request& req, highlevel_msgs::MoveTo::Response& resp)
{
    ROS_INFO("Move orientation service called: Move to x: %f, y: %f, z: %f, T: %f.", req.x, req.y, req.z, req.T);
    // Input checking
    if (req.T <= 0)
    {
        ROS_ERROR("Time for planning must be greater than 0, but entered time: %f.", req.T);
        resp.success = false;
        return false;
    }
    initial_position_ = target_position_;
    initial_ori_ = ground_truth_ori_;
    initial_T_ = ros::Time::now().toSec();

    tf2::Quaternion target_quat;
    target_quat.setRPY(req.x, req.y, req.z);
    target_ori_ = Eigen::Quaterniond(target_quat.w(), target_quat.x(), target_quat.y(), target_quat.z());
    target_T_ = initial_T_ + req.T;

    // Compute polynomial coefficients
    a0_ = 0; // Initial displacement zero
    a1_ = 0; // Initial velocity zero
    a2_ = 3 / pow(req.T, 2);
    a3_ = -2 / pow(req.T, 3);

    ROS_INFO("Request valid, trajectory execution will begin shortly.");
    resp.success = true;
    return true;
}

void CubicPolynomialPlanner::pose_callback(const geometry_msgs::Pose& msg)
{
    ground_truth_position_ << msg.position.x, msg.position.y, msg.position.z;
    ground_truth_ori_ = Eigen::Quaterniond(msg.orientation.w, msg.orientation.x, msg.orientation.y, msg.orientation.z);
    is_pose_available_ = true;
}

void CubicPolynomialPlanner::compute_trajectory_point()
{
    double current_T = ros::Time::now().toSec();
    if (current_T <= target_T_)
    {
        planning_status_msg_.data = false;
        double time_elapsed = current_T - initial_T_;
        s_ = a0_ + a1_*time_elapsed + a2_*pow(time_elapsed, 2) + a3_*pow(time_elapsed, 3);
        s_dot_ = a1_ + 2*a2_*time_elapsed + 3*a3_*pow(time_elapsed, 2);
        next_position_ = initial_position_ + s_ * (target_position_ - initial_position_);
        next_velocity_ = s_dot_ * (target_position_ - initial_position_);
        next_ori_ = initial_ori_.slerp(s_, target_ori_); 

        pose_cmd_.position.x = next_position_(0);
        pose_cmd_.position.y = next_position_(1);
        pose_cmd_.position.z = next_position_(2);
        pose_cmd_.orientation.x = next_ori_.x();
        pose_cmd_.orientation.y = next_ori_.y();
        pose_cmd_.orientation.z = next_ori_.z();
        pose_cmd_.orientation.w = next_ori_.w();
        
        twist_cmd_.linear.x = next_velocity_(0);
        twist_cmd_.linear.y = next_velocity_(1);
        twist_cmd_.linear.z = next_velocity_(2);
    } else
    {
        planning_status_msg_.data = true;
    }
}

void CubicPolynomialPlanner::update()
{
    compute_trajectory_point();
    pose_publisher_.publish(pose_cmd_);
    twist_publisher_.publish(twist_cmd_);
    planning_status_publisher_.publish(planning_status_msg_);
}