#ifndef TASK_CONTROLLER_HPP
#define TASK_CONTROLLER_HPP

#include <pinocchio/parsers/urdf.hpp>
#include <pinocchio/multibody/data.hpp>
#include <pinocchio/algorithm/jacobian.hpp>
#include <pinocchio/algorithm/kinematics.hpp>
#include <pinocchio/algorithm/compute-all-terms.hpp>
#include <pinocchio/spatial/se3.hpp>
#include <pinocchio/multibody/model.hpp>
#include <pinocchio/multibody/data.hpp>
#include <pinocchio/math/quaternion.hpp>
#include <pinocchio/math/rpy.hpp>
#include <ros/ros.h>
#include <sensor_msgs/JointState.h>
#include <highlevel_msgs/MoveTo.h>
#include <std_msgs/Float64MultiArray.h>
#include <std_msgs/Bool.h>
#include <std_srvs/Trigger.h>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/Twist.h>
#include <string>
#include <Eigen/Dense>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

class TaskController
{
public:
    TaskController(ros::NodeHandle& node_handle);
    void update();
    
private:
    bool load_params();
    bool set_initial_target();
    void joint_states_callback(const sensor_msgs::JointState& msg);
    void planner_twist_callback(const geometry_msgs::Twist& msg);
    void planner_pose_callback(const geometry_msgs::Pose& msg);
    void planner_status_callback(const std_msgs::Bool& msg);
    void compute_joint_torques();

    ros::NodeHandle node_handle_;
    ros::Publisher joint_torque_publisher_;
    ros::Publisher task_pose_publisher_;
    ros::Subscriber joint_states_subscriber_;
    ros::Subscriber planner_twist_subscriber_;
    ros::Subscriber planner_pose_subscriber_;
    ros::Subscriber planner_status_subscriber_;
    ros::ServiceClient trigger_client_;
    ros::ServiceClient planner_client_;
    geometry_msgs::Pose task_pose_msg_;
    std_msgs::Float64MultiArray joint_torque_msg_;
    highlevel_msgs::MoveTo planner_srv_;

    tf2::Quaternion quat_fbk_;
    tf2::Quaternion quat_ref_;

    pinocchio::Model robot_model_;
    pinocchio::Data robot_data_;
    pinocchio::SE3 pose_fbk_; 
    pinocchio::SE3 pose_ref_;
    pinocchio::SE3 pose_err_;
    Eigen::Quaterniond quat_init_;
    Eigen::VectorXd pose_err_vec_local_;
    Eigen::VectorXd pose_err_vec_world_;
    Eigen::VectorXd pose_ddot_cmd_;
    Eigen::VectorXd pose_dot_err_vec_;
    Eigen::VectorXd pose_dot_fbk_vec_;
    Eigen::VectorXd pose_dot_ref_vec_;
    Eigen::MatrixXd K_;
    Eigen::MatrixXd D_;
    Eigen::MatrixXd M_;
    Eigen::MatrixXd M_inv_;
    Eigen::MatrixXd h_;
    Eigen::MatrixXd j_; // Jacobian local world
    Eigen::MatrixXd j_linear_;
    Eigen::MatrixXd j_T_;
    Eigen::MatrixXd j_inv_;
    Eigen::MatrixXd j_T_inv_;
    Eigen::MatrixXd j_dot_;
    Eigen::MatrixXd j_linear_dot_;
    Eigen::MatrixXd N_;
    Eigen::VectorXd x_dot_ref_;
    Eigen::VectorXd x_ref_;
    Eigen::VectorXd x_ddot_ref_;
    Eigen::VectorXd q_ref_;
    Eigen::VectorXd q_dot_ref_;
    Eigen::VectorXd q_fbk_;
    Eigen::VectorXd q_dot_fbk_;
    Eigen::VectorXd x_fbk_;
    Eigen::VectorXd x_dot_fbk_;
    Eigen::VectorXd tau_null_;
    Eigen::VectorXd F_; // Wrench
    Eigen::MatrixXd lambda_; 
    Eigen::VectorXd eta_;
    Eigen::VectorXd tau_;
    Eigen::MatrixXd P_;
    Eigen::VectorXd q0_;
    Eigen::VectorXd q_ddot_cmd_;

    std::string joint_states_topic_;
    std::string joint_publisher_topic_;
    std::string urdf_path_;
    std::vector<double> initial_position_list_;
    std::vector<double> initial_orientation_list_;
    std::vector<double> joint_target_list_;
    std::vector<double> joint_velocity_limits_;
    double initial_T_;
    double loop_rate_;
    double loop_time_;
    double z_min_;
    double K_trans_;
    double D_trans_;
    double K_ori_;
    double D_ori_;
    double K_q_;
    double D_q_;
    double k_err_dot_;
    int n_joints_;
    bool is_planner_done_;
    bool is_initialized_;
    bool is_pose_detected_;

    const double epsilon_ = 1e-6; // Small value to avoid division by zero
};

#endif // TASK_CONTROLLER_HPP