#include "dynamic_controller/JointController.hpp"
#include <std_srvs/Trigger.h>

JointController::JointController(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    ROS_INFO("Started node planner");
    if (!load_params())
    {
        ROS_ERROR("Shutting down thread");
        ros::requestShutdown();
    }
    loop_time_ = 1 / loop_rate_;

    joint_torque_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>(torque_cmd_topic_, 1, this);
    joint_states_subscriber_ = node_handle_.subscribe(joint_states_topic_, 1, &JointController::joint_states_callback, this);
    velocity_cmd_subscriber_ = node_handle_.subscribe("/planner/joint_velocities", 1, &JointController::velocity_cmd_callback, this);
    position_cmd_subscriber_ = node_handle_.subscribe("/planner/joint_positions", 1, &JointController::position_cmd_callback, this);
    planning_status_subscriber_ = node_handle_.subscribe("/planner/done", 1, &JointController::planning_status_callback, this);
    trigger_client_ = node_handle_.serviceClient<std_srvs::Trigger>("/planner/start", this);

    pinocchio::urdf::buildModel(urdf_path_, robot_model_, false);
    robot_data_ = pinocchio::Data(robot_model_);
    n_joints_ = robot_model_.nq;

    // Initialize Eigen
    M_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_);
    h_ = Eigen::MatrixXd::Zero(n_joints_, 1);
    q_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    qd_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    q_ref_ = Eigen::VectorXd::Zero(n_joints_);
    qd_ref_ = Eigen::VectorXd::Zero(n_joints_);
    qdd_ref_ = Eigen::VectorXd::Zero(n_joints_);
    qdd_cmd_ = Eigen::VectorXd::Zero(n_joints_);
    tau_cmd_ = Eigen::VectorXd::Zero(n_joints_);

    joint_torque_cmd_.data.resize(n_joints_);
    is_target_reached_ = false;
    is_joint_detected_ = false;
    std_srvs::Trigger trigger_srv;
    if (trigger_client_.call(trigger_srv))
    {
        ROS_INFO("Calling /planner/start");
    } else
    {
        ROS_ERROR("Failed to call /planner/start");
        ros::requestShutdown();
    }
}

bool JointController::load_params()
{
    if(!node_handle_.getParam("/publish_rate", loop_rate_))
    {
        ROS_ERROR("Could not load parameter /publish_rate");
        return false;
    }
    if(!node_handle_.getParam("K", K_))
    {
        ROS_ERROR("Could not retrieve parameter K");
        return false;
    }
    if(!node_handle_.getParam("D", D_))
    {
        ROS_ERROR("Could not retrieve parameter D");
        return false;
    }
    if(!node_handle_.getParam("urdf_file_name", urdf_path_))
    {
        ROS_ERROR("Could not retrieve parameter urdf_file_name");
        return false;
    }
    if(!node_handle_.getParam("torque_cmd_topic", torque_cmd_topic_))
    {
        ROS_ERROR("Could not retrieve parameter torque_cmd_topic");
        return false;
    }
    if(!node_handle_.getParam("joint_states_topic", joint_states_topic_))
    {
        ROS_ERROR("Could not retrieve parameter joint_states_topic");
        return false;
    }
    if (!node_handle_.getParam("zero_vel", is_zero_vel_)) 
    {
        ROS_ERROR("Could not retrieve parameter zero_vel");
        return false;
    }
    ROS_INFO("Parameters loaded succesfully");
    return true;
}

void JointController::joint_states_callback(const sensor_msgs::JointState& msg)
{
    for (int i = 0; i < n_joints_; ++i)
    {
        q_fbk_(i) = msg.position[i];
        qd_fbk_(i) = msg.velocity[i];
    }
    is_joint_detected_ = true;
}

void JointController::velocity_cmd_callback(const std_msgs::Float64MultiArray& msg)
{
    if (is_zero_vel_) return; // If zero_vel is true, do not update qd_ref_
    for (int i = 0; i < n_joints_; ++i)
    {
        qd_ref_(i) = msg.data[i];
    }
}

void JointController::position_cmd_callback(const std_msgs::Float64MultiArray& msg)
{
    // if(!is_target_reached_)
    // {
    //     joint_position_cmd_ = msg;
    // }
    for (int i = 0; i < n_joints_; ++i)
    {
        q_ref_(i) = msg.data[i];
    }
}

void JointController::planning_status_callback(const std_msgs::Bool& msg)
{
    is_target_reached_ = msg.data;
}

void JointController::compute_joint_torques()
{
    qdd_cmd_ = qdd_ref_ + K_ * (q_ref_ - q_fbk_) + D_ * (qd_ref_ - qd_fbk_);
    //ROS_DEBUG_STREAM("V: " << D_ * (qd_ref_ - qd_fbk_));
    pinocchio::computeAllTerms(robot_model_, robot_data_, q_fbk_, qd_fbk_);
    M_ = robot_data_.M;
    h_ = robot_data_.nle;
    tau_cmd_ = M_ * qdd_cmd_ + h_;
    for (int i = 0; i < n_joints_; ++i)
    {
        joint_torque_cmd_.data[i] = tau_cmd_(i);
    }
}

void JointController::update()
{
    if (is_joint_detected_)
    {
        compute_joint_torques();
    }
    joint_torque_publisher_.publish(joint_torque_cmd_);
}