#include <dynamic_controller/TaskController.hpp>
#include <Eigen/QR>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

TaskController::TaskController(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    if(!load_params())
    {
        ROS_ERROR("Shutting down thread");
        ros::requestShutdown();
    }
    loop_time_ = 1 / loop_rate_;
    joint_torque_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>(joint_publisher_topic_, 1, this);
    task_pose_publisher_ = node_handle_.advertise<geometry_msgs::Pose>("pose", 1, this);
    planner_twist_subscriber_ = node_handle_.subscribe("/pose_planner/twist_cmd", 1, &TaskController::planner_twist_callback, this);
    planner_pose_subscriber_ = node_handle_.subscribe("/pose_planner/pose_cmd", 1, &TaskController::planner_pose_callback, this);
    planner_status_subscriber_ = node_handle_.subscribe("/pose_planner/done", 1, &TaskController::planner_status_callback, this);
    joint_states_subscriber_ = node_handle_.subscribe(joint_states_topic_, 1, &TaskController::joint_states_callback, this);
    planner_client_ = node_handle_.serviceClient<highlevel_msgs::MoveTo>("/pose_planner/move_to", this);
    trigger_client_ = node_handle_.serviceClient<std_srvs::Trigger>("/planner/start", this);

    // Pinocchio
    pinocchio::urdf::buildModel(urdf_path_, robot_model_, false);
    robot_data_ = pinocchio::Data(robot_model_);
    n_joints_ = robot_model_.nq;
    j_ = Eigen::MatrixXd::Zero(6, n_joints_); // Jacobian full
    j_linear_ = Eigen::MatrixXd::Zero(3, n_joints_); // Linear velocities only
    j_T_ = Eigen::MatrixXd::Zero(n_joints_, 3); // Transpose of jacobian
    j_dot_ = Eigen::MatrixXd::Zero(6, n_joints_); // Time derivative of jacobian
    j_linear_dot_ = Eigen::MatrixXd::Zero(3, n_joints_); // Linear accelerations only
    j_inv_ = Eigen::MatrixXd::Zero(n_joints_, 3); // Pseudo-inverse of jacobian
    j_T_inv_ = Eigen::MatrixXd::Zero(3, n_joints_); // Pseudo-inverse of jacobian transpose
    N_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_); 

    pose_fbk_ = pinocchio::SE3::Identity();
    pose_ref_ = pinocchio::SE3::Identity();
    pose_err_ = pinocchio::SE3::Identity();
    pose_err_vec_local_ = Eigen::VectorXd::Zero(6);
    pose_err_vec_world_ = Eigen::VectorXd::Zero(6);
    pose_ddot_cmd_ = Eigen::VectorXd::Zero(6);
    pose_dot_err_vec_ = Eigen::VectorXd::Zero(6);
    pose_dot_fbk_vec_ = Eigen::VectorXd::Zero(6);
    pose_dot_ref_vec_ = Eigen::VectorXd::Zero(6);

    K_ = Eigen::MatrixXd::Zero(6, 6);
    D_ = Eigen::MatrixXd::Zero(6, 6);
    K_.topLeftCorner(3, 3) = K_trans_ * Eigen::MatrixXd::Identity(3, 3);
    K_.bottomRightCorner(3, 3) = K_ori_ * Eigen::MatrixXd::Identity(3, 3);
    D_.topLeftCorner(3, 3) = D_trans_ * Eigen::MatrixXd::Identity(3, 3);
    D_.bottomRightCorner(3, 3) = D_ori_ * Eigen::MatrixXd::Identity(3, 3);


    x_ddot_ref_ = Eigen::VectorXd::Zero(3);
    x_dot_ref_ = Eigen::VectorXd::Zero(3);
    x_ref_ = Eigen::VectorXd::Zero(3);
    q_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    q_ref_ = q_fbk_;
    q_dot_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    q_dot_ref_ = q_dot_fbk_;
    x_fbk_ = Eigen::VectorXd::Zero(3);
    x_dot_fbk_ = Eigen::VectorXd::Zero(3);
    M_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_);
    M_inv_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_);
    h_ = Eigen::MatrixXd::Zero(n_joints_, 1);
    F_ = Eigen::VectorXd::Zero(6);
    lambda_ = Eigen::MatrixXd::Zero(6, 6);
    eta_ = Eigen::VectorXd::Zero(6);
    tau_ = Eigen::VectorXd::Zero(n_joints_);
    tau_null_ = Eigen::VectorXd::Zero(n_joints_);
    P_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_);
    q0_ = Eigen::VectorXd::Zero(n_joints_);
    for (int i = 0; i < n_joints_; ++i)
    {
        q0_(i) = joint_target_list_[i];
    }
    Eigen::Matrix3d rot_init = pinocchio::rpy::rpyToMatrix(initial_orientation_list_[0], initial_orientation_list_[1], initial_orientation_list_[2]);
    pinocchio::quaternion::assignQuaternion(quat_init_, rot_init);

    is_planner_done_ = false;
    is_initialized_ = false;
    is_pose_detected_ = false;
    joint_torque_msg_.data.resize(n_joints_);
}

bool TaskController::load_params()
{
    // Get parameters from parameter server
    if(!node_handle_.getParam("/publish_rate", loop_rate_))
    {
        ROS_ERROR("Could not load parameter /publish_rate");
        return false;
    }
    if(!node_handle_.getParam("initial_position_goal", initial_position_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter initial_position_goal");
        return false;
    }
    if (!node_handle_.getParam("initial_orientation_goal", initial_orientation_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter initial_orientation_goal");
        return false;
    }
    if (!node_handle_.getParam("joint_target_list", joint_target_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter joint_target_list");
        return false;
    }
    if(!node_handle_.getParam("joint_states_topic", joint_states_topic_))
    {
        ROS_ERROR("Could not load parameter pose_subscriber_topic");
        return false;
    }
    if(!node_handle_.getParam("joint_publisher_topic", joint_publisher_topic_))
    {
        ROS_ERROR("Could not load parameter joint_publisher_topic");
        return false;
    }
    if (!node_handle_.getParam("urdf_file_name", urdf_path_)) 
    {
        ROS_ERROR("Could not retrieve parameter urdf_file_name");
        return false;
    }
    if (!node_handle_.getParam("initial_T", initial_T_))
    {
        ROS_ERROR("Could not retrieve parameter initial_T");
        return false;
    }
    if (!node_handle_.getParam("K_trans", K_trans_))
    {
        ROS_ERROR("Could not retrieve parameter K_trans");
        return false;
    }
    if (!node_handle_.getParam("D_trans", D_trans_))
    {
        ROS_ERROR("Could not retrieve parameter D_trans");
        return false;
    }
    if (!node_handle_.getParam("K_q", K_q_))
    {
        ROS_ERROR("Could not retrieve parameter K_q");
        return false;
    }
    if (!node_handle_.getParam("D_q", D_q_))
    {
        ROS_ERROR("Could not retrieve parameter D_q");
        return false;
    }
        if (!node_handle_.getParam("K_ori", K_ori_))
    {
        ROS_ERROR("Could not retrieve parameter K_ori");
        return false;
    }
    if (!node_handle_.getParam("D_ori", D_ori_))
    {
        ROS_ERROR("Could not retrieve parameter D_ori");
        return false;
    }
    if (!node_handle_.getParam("k_err_dot", k_err_dot_))
    {
        ROS_ERROR("Could not retrieve parameter k_err_dot");
        return false;
    }

    return true;
}

bool TaskController::set_initial_target()
{
    planner_srv_.request.x = initial_position_list_[0];
    planner_srv_.request.y = initial_position_list_[1];
    planner_srv_.request.z = initial_position_list_[2];
    planner_srv_.request.T = initial_T_;
    if (planner_client_.call(planner_srv_))
    {
        ROS_INFO("Moving to start position");
    } else
    {
        ROS_INFO("Try again");
    }
    return planner_srv_.response.success;
}

void TaskController::joint_states_callback(const sensor_msgs::JointState& msg)
{
    for (int i = 0; i < n_joints_; ++i)
    {
        q_fbk_(i) = msg.position[i];
        q_dot_fbk_(i) = msg.velocity[i];
    }

    //pinocchio::forwardKinematics(robot_model_, robot_data_, q_fbk_);
    pinocchio::computeAllTerms(robot_model_, robot_data_, q_fbk_, q_dot_fbk_);
    pose_fbk_ = robot_data_.oMi[n_joints_];
    pinocchio::getJointJacobian(robot_model_, robot_data_, n_joints_, pinocchio::ReferenceFrame::LOCAL_WORLD_ALIGNED, j_);
    pinocchio::getJointJacobianTimeVariation(robot_model_, robot_data_, n_joints_, pinocchio::ReferenceFrame::LOCAL_WORLD_ALIGNED, j_dot_);
    j_linear_dot_ = j_dot_.topRows(3);
    j_linear_ = j_.topRows(3);
    x_dot_fbk_ = j_linear_ * q_dot_fbk_;
    pose_dot_fbk_vec_ = j_ * q_dot_fbk_;
    x_fbk_ = pose_fbk_.translation();
    task_pose_msg_.position.x = x_fbk_(0);
    task_pose_msg_.position.y = x_fbk_(1);
    task_pose_msg_.position.z = x_fbk_(2);
    Eigen::Quaterniond q;
    if (!is_initialized_)
    {
        q = quat_init_; // Set so planner publishes correct orientation
    } else
    {
        pinocchio::quaternion::assignQuaternion(q, pose_fbk_.rotation());
    }
    task_pose_msg_.orientation.x = q.x();
    task_pose_msg_.orientation.y = q.y();
    task_pose_msg_.orientation.z = q.z();
    task_pose_msg_.orientation.w = q.w();
    tf2::fromMsg(task_pose_msg_.orientation, quat_fbk_);
    is_pose_detected_ = true; 
}

void TaskController::planner_twist_callback(const geometry_msgs::Twist& msg)
{
    x_dot_ref_ << msg.linear.x, msg.linear.y, msg.linear.z;
    pose_dot_ref_vec_.head<3>() = x_dot_ref_;
}

void TaskController::planner_pose_callback(const geometry_msgs::Pose& msg)
{
    x_ref_ << msg.position.x, msg.position.y, msg.position.z;
    pose_ref_.translation() = x_ref_;
    Eigen::Quaterniond q(msg.orientation.w, msg.orientation.x, msg.orientation.y, msg.orientation.z);
    pose_ref_.rotation() = q.toRotationMatrix();
    tf2::fromMsg(msg.orientation, quat_ref_);
}


void TaskController::planner_status_callback(const std_msgs::Bool& msg)
{
    if (msg.data && !is_planner_done_)
    {
        ROS_INFO("Trajectory completed");
    }
    is_planner_done_ = msg.data;
}

void TaskController::compute_joint_torques()
{
    pose_err_ = pose_fbk_.inverse() * pose_ref_;
    pose_err_vec_local_ = pinocchio::log6(pose_err_).toVector();
    pose_err_vec_world_.head<3>() = x_ref_ - x_fbk_;
    pose_err_vec_world_.tail<3>() = pose_fbk_.rotation() * pose_err_vec_local_.tail<3>();
    pose_dot_ref_vec_.tail<3>() = pose_err_vec_world_.tail<3>() / loop_time_ * k_err_dot_;
    pose_dot_err_vec_ = pose_dot_ref_vec_ - pose_dot_fbk_vec_;
    pose_ddot_cmd_ = K_ * (pose_err_vec_world_) + D_ * (pose_dot_err_vec_);

    j_T_ = j_.transpose();
    Eigen::MatrixXd j_T_inv_reg = Eigen::MatrixXd::Identity(n_joints_, n_joints_) * epsilon_;
    Eigen::MatrixXd j_inv_reg = Eigen::MatrixXd::Identity(6, 6) * epsilon_;
    j_inv_ = j_T_ * (j_ * j_T_ + j_inv_reg).inverse();
    j_T_inv_ = j_ * (j_T_ * j_ + j_T_inv_reg).inverse();
    M_ = robot_data_.M;
    M_inv_ = M_.inverse();
    h_ = robot_data_.nle;
    lambda_ = j_T_inv_ * M_ * j_inv_;
    eta_ = j_T_inv_*h_ + lambda_*j_dot_*q_dot_fbk_; 
    F_ = lambda_ * pose_ddot_cmd_ + eta_;
    tau_ = j_T_ * F_;

    q_ddot_cmd_ = K_q_ * (q0_ - q_fbk_) + D_q_ * (q_dot_ref_ - q_dot_fbk_);
    tau_null_ = M_ * q_ddot_cmd_ + h_;
    P_ = Eigen::MatrixXd::Identity(n_joints_, n_joints_) - j_T_ * (j_ * M_inv_ * j_T_).inverse() * j_ * M_inv_;
    tau_ = tau_ + P_ * tau_null_;
    for (int i = 0; i < n_joints_; ++i)
    {
        joint_torque_msg_.data[i] = tau_(i);
    }
}

void TaskController::update()
{
    if (is_pose_detected_)
    {
        task_pose_publisher_.publish(task_pose_msg_);
    }
    if (!is_initialized_)
    {
        if (!set_initial_target()) return;
        is_initialized_ = true;
    }
    compute_joint_torques();
    joint_torque_publisher_.publish(joint_torque_msg_);
    // ROS_INFO_STREAM("Error: " << (x_fbk_-x_ref_).norm());
    // ROS_INFO_STREAM("Shortest angle: " << tf2::angleShortestPath(quat_fbk_, quat_ref_));
}