#include "highlevel_controller/PoseActionClient.hpp"
#include <highlevel_msgs/PoseCommandAction.h>

PoseActionClient::PoseActionClient(ros::NodeHandle& node_handle) : node_handle_(node_handle), action_client_("pose_command", true)
{
    if (!load_params()) {
        ROS_ERROR("Could not load parameters!");
        ros::requestShutdown();
    }
    target_index_ = 0;
}

bool PoseActionClient::load_params() {
    if (!node_handle_.getParam("feedback_subscriber_topic", feedback_subscriber_topic_)) {
        ROS_ERROR("Could not retrieve feedback_subscriber_topic parameter!");
        return false;
    }
    if (!node_handle_.getParam("/action_list/number_of_targets", n_targets_)) {
        ROS_ERROR("Could not retrieve /action_list/number_of_targets parameter!");
        return false;
    }
    target_orientations_ = Eigen::MatrixXd::Zero(n_targets_, 3);
    target_translations_ = Eigen::MatrixXd::Zero(n_targets_, 3);
    target_Ts_ = Eigen::VectorXd::Zero(n_targets_);

    for (int i = 0; i < n_targets_; ++i)
    {
        std::string target_name = "/action_list/action_" + std::to_string(i);
        std::vector<double> target_translation;
        std::vector<double> target_orientation;
        double target_T;
        if (!node_handle_.getParam(target_name + "/translation", target_translation)) 
        {
            ROS_ERROR_STREAM("Could not retrieve " << target_name << "/translation parameter!");
            return false;
        }
        if (!node_handle_.getParam(target_name + "/orientation", target_orientation)) 
        {
            ROS_ERROR_STREAM("Could not retrieve " << target_name << "/orientation parameter!");
            return false;
        }
        if (!node_handle_.getParam(target_name + "/duration", target_T))
        {
            ROS_ERROR_STREAM("Could not retrieve " << target_name << "/duration parameter!");
            return false;
        }
        target_translations_.row(i) << target_translation[0], target_translation[1], target_translation[2];
        target_orientations_.row(i) << target_orientation[0], target_orientation[1], target_orientation[2];
        target_Ts_(i) = target_T;
    }
    return true;
}

void PoseActionClient::update()
{
    for (int i = 0; i < n_targets_; ++i)
    {
        action_goal_.x = target_translations_(i, 0);
        action_goal_.y = target_translations_(i, 1);
        action_goal_.z = target_translations_(i, 2);
        action_goal_.roll = target_orientations_(i, 0);
        action_goal_.pitch = target_orientations_(i, 1);
        action_goal_.yaw = target_orientations_(i, 2);
        action_goal_.T = target_Ts_(i);

        ROS_INFO_STREAM("Waiting for action server to start.");
        action_client_.waitForServer();
        ROS_INFO_STREAM("Sending goal " << i << " to action server");
        action_client_.sendGoal(action_goal_,
            boost::bind(&PoseActionClient::done_callback, this, _1, _2),
            boost::bind(&PoseActionClient::active_callback, this),
            boost::bind(&PoseActionClient::feedback_callback, this, _1));

        action_client_.waitForResult(ros::Duration(30.0));
        if (action_client_.getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
        {
            ROS_INFO_STREAM("Action " << i << " succeeded!");
        }
    }
    ROS_INFO_STREAM("Action client finished!");
    ros::shutdown();
}

void PoseActionClient::done_callback(const actionlib::SimpleClientGoalState& state, const highlevel_msgs::PoseCommandResultConstPtr& result)
{
    // ROS_INFO_STREAM("Action finished in state " << state.toString().c_str());
    ROS_INFO_STREAM("Distance trans: " << distance_translation_ << ", distance ori: " << distance_orientation_ << ", time elapsed: " << time_elapsed_);
}

void PoseActionClient::active_callback()
{
    ROS_INFO_STREAM("Action has become active!");
}

void PoseActionClient::feedback_callback(const highlevel_msgs::PoseCommandFeedbackConstPtr& feedback)
{
    // ROS_INFO_STREAM("Feedback received: trans: " << feedback->distance_translation << ", ori: " << feedback->distance_orientation << ", t: " << feedback->time_elapsed);
    distance_translation_ = feedback->distance_translation;
    distance_orientation_ = feedback->distance_orientation;
    time_elapsed_ = feedback->time_elapsed;
}
