#ifndef TASK_CONTROLLER_HPP
#define TASK_CONTROLLER_HPP

#include <pinocchio/multibody/model.hpp>
#include <pinocchio/multibody/data.hpp>
#include <pinocchio/algorithm/jacobian.hpp>
#include <pinocchio/algorithm/kinematics.hpp>
#include <pinocchio/algorithm/compute-all-terms.hpp>
#include <pinocchio/spatial/se3.hpp>
#include <pinocchio/math/quaternion.hpp>
#include <pinocchio/math/rpy.hpp>
#include <ros/ros.h>
#include <sensor_msgs/JointState.h>
#include <highlevel_msgs/MoveTo.h>
#include <std_msgs/Float64MultiArray.h>
#include <std_msgs/Bool.h>
#include <std_srvs/Trigger.h>
#include <geometry_msgs/Pose.h>
#include <geometry_msgs/Twist.h>
#include <string>
#include <Eigen/Dense>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

class TaskController
{
public:
    TaskController(ros::NodeHandle& node_handle);
    void update();
    
private:
    bool load_params();
    bool set_initial_target();
    void start_potential_field();
    void joint_states_callback(const sensor_msgs::JointState& msg);
    void planner_twist_callback(const geometry_msgs::Twist& msg);
    void planner_pose_callback(const geometry_msgs::Pose& msg);
    void planner_status_callback(const std_msgs::Bool& msg);
    void potential_velocity_callback(const std_msgs::Float64MultiArray& msg);
    void compute_joint_positions();

    ros::NodeHandle node_handle_;
    ros::Publisher joint_position_publisher_;
    ros::Publisher task_pose_publisher_;
    ros::Subscriber joint_states_subscriber_;
    ros::Subscriber planner_twist_subscriber_;
    ros::Subscriber planner_pose_subscriber_;
    ros::Subscriber planner_status_subscriber_;
    ros::Subscriber potential_velocity_subscriber_;
    ros::ServiceClient trigger_client_;
    ros::ServiceClient planner_client_;
    geometry_msgs::Pose task_pose_msg_;
    std_msgs::Float64MultiArray joint_position_msg_;
    highlevel_msgs::MoveTo planner_srv_;

    pinocchio::Model robot_model_;
    pinocchio::Data robot_data_;
    pinocchio::SE3 pose_fbk_;
    pinocchio::SE3 pose_ref_;
    pinocchio::SE3 pose_err_;
    Eigen::VectorXd pose_err_vec_local_;
    // Eigen::VectorXd pose_err_vec_world_;
    Eigen::VectorXd pose_dot_ref_cmd_;
    Eigen::Quaterniond quat_init_;
    tf2::Quaternion quat_ref_;
    tf2::Quaternion quat_fbk_;

    Eigen::MatrixXd j_;
    Eigen::MatrixXd jacobian_linear_;
    Eigen::MatrixXd j_inv_;
    Eigen::MatrixXd nullspace_matrix_;
    Eigen::VectorXd x_dot_ref_;
    Eigen::VectorXd x_ref_;
    Eigen::VectorXd q_cmd_;
    Eigen::VectorXd q_dot_cmd_;
    Eigen::VectorXd next_joint_velocity_;
    Eigen::VectorXd q_fbk_;
    Eigen::VectorXd q_dot_fbk_;
    Eigen::VectorXd x_fbk_;
    Eigen::VectorXd q_dot_null_;

    std::string joint_states_topic_;
    std::string joint_publisher_topic_;
    std::string urdf_path_;
    std::vector<double> initial_position_list_;
    std::vector<double> initial_orientation_list_;
    std::vector<double> joint_target_list_;
    std::vector<double> joint_velocity_limits_;
    double initial_T_;
    double loop_rate_;
    double loop_time_;
    double z_min_;
    int n_joints_;
    bool is_planner_done_;
    bool is_initialized_;
    bool is_pose_detected_;
};

#endif // TASK_CONTROLLER_HPP