#include "kinematic_controller/JointController.hpp"
#include <std_srvs/Trigger.h>

JointController::JointController(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    ROS_INFO("Started node planner");
    if (!load_params())
    {
        ROS_ERROR("Shutting down thread");
        ros::requestShutdown();
    }
    joint_velocity_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>(velocity_cmd_topic_, 1, this);
    joint_position_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>(position_cmd_topic_, 1, this);
    velocity_cmd_subscriber_ = node_handle_.subscribe("/planner/joint_velocities", 1, &JointController::velocity_cmd_callback, this);
    position_cmd_subscriber_ = node_handle_.subscribe("/planner/joint_positions", 1, &JointController::position_cmd_callback, this);
    planning_status_subscriber_ = node_handle_.subscribe("/planner/done", 1, &JointController::planning_status_callback, this);
    trigger_client_ = node_handle_.serviceClient<std_srvs::Trigger>("/planner/start", this);
    joint_velocity_cmd_.data.resize(n_joints_);
    joint_position_cmd_.data.resize(n_joints_);
    is_target_reached_ = false;
    std_srvs::Trigger trigger_srv;
    if (trigger_client_.call(trigger_srv))
    {
        ROS_INFO("Calling /planner/start");
    } else
    {
        ROS_ERROR("Failed to call /planner/start");
        ros::requestShutdown();
    }
}

bool JointController::load_params()
{
    if(!node_handle_.getParam("position_cmd_topic", position_cmd_topic_))
    {
        ROS_ERROR("Could not retrieve parameter position_cmd_topic");
        return false;
    }
    if (!node_handle_.getParam("velocity_cmd_topic", velocity_cmd_topic_))
    {
        ROS_ERROR("Could not retrieve parameter velocity_cmd_topic");
        return false;
    }
    if(!node_handle_.getParam("n_joints", n_joints_))
    {
        ROS_ERROR("Could not retrieve parameter n_joints");
        return false;
    }
    ROS_INFO("Parameters loaded succesfully");
    return true;
}

void JointController::velocity_cmd_callback(const std_msgs::Float64MultiArray& msg)
{
    if(!is_target_reached_)
    {
        joint_velocity_cmd_ = msg;
    }
}

void JointController::position_cmd_callback(const std_msgs::Float64MultiArray& msg)
{
    if(!is_target_reached_)
    {
        joint_position_cmd_ = msg;
    }
}

void JointController::planning_status_callback(const std_msgs::Bool& msg)
{
    if (msg.data && !is_target_reached_)
    {
        ROS_INFO("Reached joint targets");
    }
    is_target_reached_ = msg.data;
}

void JointController::update()
{
    joint_velocity_publisher_.publish(joint_velocity_cmd_);
    joint_position_publisher_.publish(joint_position_cmd_);
}