#include <pinocchio/parsers/urdf.hpp>
#include <pinocchio/multibody/data.hpp>
#include <pinocchio/algorithm/jacobian.hpp>
#include <pinocchio/algorithm/kinematics.hpp>
#include <pinocchio/algorithm/compute-all-terms.hpp>
#include <pinocchio/spatial/se3.hpp>
#include "kinematic_controller/TaskController.hpp"

TaskController::TaskController(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    if(!load_params())
    {
        ROS_ERROR("Shutting down thread");
        ros::requestShutdown();
    }
    loop_time_ = 1 / loop_rate_;
    joint_position_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>(joint_publisher_topic_, 1, this);
    task_pose_publisher_ = node_handle_.advertise<geometry_msgs::Pose>("pose", 1, this);
    //planner_twist_subscriber_ = node_handle_.subscribe("/pose_planner/twist_cmd", 1, &TaskController::planner_twist_callback, this);
    planner_pose_subscriber_ = node_handle_.subscribe("/pose_planner/pose_cmd", 1, &TaskController::planner_pose_callback, this);
    planner_status_subscriber_ = node_handle_.subscribe("/pose_planner/done", 1, &TaskController::planner_status_callback, this);
    joint_states_subscriber_ = node_handle_.subscribe(joint_states_topic_, 1, &TaskController::joint_states_callback, this);
    potential_velocity_subscriber_ = node_handle_.subscribe("/planner/joint_velocities", 1, &TaskController::potential_velocity_callback, this);
    planner_client_ = node_handle_.serviceClient<highlevel_msgs::MoveTo>("/pose_planner/move_to", this);
    trigger_client_ = node_handle_.serviceClient<std_srvs::Trigger>("/planner/start", this);

    // Pinocchio
    pinocchio::urdf::buildModel(urdf_path_, robot_model_, false);
    robot_data_ = pinocchio::Data(robot_model_);
    n_joints_ = robot_model_.nq;
    j_ = Eigen::MatrixXd::Zero(6, n_joints_); // Jacobian full
    jacobian_linear_ = Eigen::MatrixXd::Zero(3, n_joints_); // Linear velocities only
    j_inv_ = Eigen::MatrixXd::Zero(n_joints_, 6); // Pseudo-inverse of jacobian
    nullspace_matrix_ = Eigen::MatrixXd::Zero(n_joints_, n_joints_);
    Eigen::MatrixXd rot_init = pinocchio::rpy::rpyToMatrix(initial_orientation_list_[0], initial_orientation_list_[1], initial_orientation_list_[2]);
    pinocchio::quaternion::assignQuaternion(quat_init_, rot_init);
    pose_err_vec_local_ = Eigen::VectorXd::Zero(6);
    pose_dot_ref_cmd_ = Eigen::VectorXd::Zero(6);

    x_dot_ref_ = Eigen::VectorXd::Zero(3);
    x_ref_ = Eigen::VectorXd::Zero(3);
    q_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    q_cmd_ = q_fbk_;
    q_dot_fbk_ = Eigen::VectorXd::Zero(n_joints_);
    q_dot_cmd_ = q_dot_fbk_;
    q_dot_null_ = q_dot_fbk_;
    x_fbk_ = Eigen::VectorXd::Zero(3);
    is_planner_done_ = false;
    is_initialized_ = false;
    is_pose_detected_ = false;
    joint_position_msg_.data.resize(n_joints_);
}

bool TaskController::load_params()
{
    // Get parameters from parameter server
    if(!node_handle_.getParam("/publish_rate", loop_rate_))
    {
        ROS_ERROR("Could not load parameter /publish_rate");
        return false;
    }
    if(!node_handle_.getParam("initial_position_goal", initial_position_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter initial_position_goal");
        return false;
    }
    if (!node_handle_.getParam("initial_orientation_goal", initial_orientation_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter initial_orientation_goal");
        return false;
    }
    if (!node_handle_.getParam("joint_target_list", joint_target_list_)) 
    {
        ROS_ERROR("Could not retrieve parameter joint_target_list");
        return false;
    }
    if(!node_handle_.getParam("joint_states_topic", joint_states_topic_))
    {
        ROS_ERROR("Could not load parameter pose_subscriber_topic");
        return false;
    }
    if(!node_handle_.getParam("joint_publisher_topic", joint_publisher_topic_))
    {
        ROS_ERROR("Could not load parameter joint_publisher_topic");
        return false;
    }
    if (!node_handle_.getParam("urdf_file_name", urdf_path_)) 
    {
        ROS_ERROR("Could not retrieve parameter urdf_file_name");
        return false;
    }
    if (!node_handle_.getParam("initial_T", initial_T_))
    {
        ROS_ERROR("Could not retrieve parameter initial_T");
        return false;
    }
    return true;
}

bool TaskController::set_initial_target()
{
    planner_srv_.request.x = initial_position_list_[0];
    planner_srv_.request.y = initial_position_list_[1];
    planner_srv_.request.z = initial_position_list_[2];
    planner_srv_.request.T = initial_T_;
    if (planner_client_.call(planner_srv_))
    {
        ROS_INFO("Moving to start position");
    } else
    {
        ROS_INFO("Try again");
    }
    return planner_srv_.response.success;
}

void TaskController::start_potential_field()
{
    std_srvs::Trigger trigger_srv;
    if (trigger_client_.call(trigger_srv))
    {
        ROS_INFO("Calling /planner/start");
    } else
    {
        ROS_ERROR("Failed to call /planner/start");
        ros::requestShutdown();
    }
}

void TaskController::joint_states_callback(const sensor_msgs::JointState& msg)
{
    for (int i = 0; i < n_joints_; ++i)
    {
        q_fbk_(i) = msg.position[i];
        q_dot_fbk_(i) = msg.velocity[i];
    }
    pinocchio::forwardKinematics(robot_model_, robot_data_, q_fbk_);
    pose_fbk_ = robot_data_.oMi[n_joints_];
    x_fbk_ = pose_fbk_.translation();
    task_pose_msg_.position.x = x_fbk_(0);
    task_pose_msg_.position.y = x_fbk_(1);
    task_pose_msg_.position.z = x_fbk_(2);
    Eigen::Quaterniond q;
    pinocchio::quaternion::assignQuaternion(q, pose_fbk_.rotation());
    task_pose_msg_.orientation.x = q.x();
    task_pose_msg_.orientation.y = q.y();
    task_pose_msg_.orientation.z = q.z();
    task_pose_msg_.orientation.w = q.w();
    tf2::fromMsg(task_pose_msg_.orientation, quat_fbk_);
    is_pose_detected_ = true; 
}

void TaskController::planner_twist_callback(const geometry_msgs::Twist& msg)
{
    x_dot_ref_ << msg.linear.x, msg.linear.y, msg.linear.z;
}

void TaskController::planner_pose_callback(const geometry_msgs::Pose& msg)
{
    x_ref_ << msg.position.x, msg.position.y, msg.position.z;
    pose_ref_.translation() = x_ref_;
    Eigen::Quaterniond q(msg.orientation.w, msg.orientation.x, msg.orientation.y, msg.orientation.z);
    pose_ref_.rotation() = q.toRotationMatrix();
    tf2::fromMsg(msg.orientation, quat_ref_);
    if (!is_initialized_)
    {
        is_initialized_ = true;
        start_potential_field();
    }
}

void TaskController::planner_status_callback(const std_msgs::Bool& msg)
{
    if (msg.data && !is_planner_done_)
    {
        ROS_INFO("Trajectory completed");
    }
    is_planner_done_ = msg.data;
}

void TaskController::potential_velocity_callback(const std_msgs::Float64MultiArray& msg)
{
    for (int i = 0; i < n_joints_; ++i)
    {
        q_dot_null_(i) = msg.data[i];
    }
}

void TaskController::compute_joint_positions()
{
    pinocchio::computeAllTerms(robot_model_, robot_data_, q_fbk_, q_dot_fbk_);
    pinocchio::getJointJacobian(robot_model_, robot_data_, n_joints_, pinocchio::ReferenceFrame::LOCAL_WORLD_ALIGNED, j_);
    pose_err_ = pose_fbk_.inverse() * pose_ref_;
    pose_err_vec_local_ = pinocchio::log6(pose_err_).toVector();
    x_dot_ref_ = (x_ref_ - x_fbk_) / loop_time_;
    pose_dot_ref_cmd_.head<3>() = x_dot_ref_;
    pose_dot_ref_cmd_.tail<3>() = pose_fbk_.rotation() * pose_err_vec_local_.tail<3>() / loop_time_;
    j_inv_ = j_.transpose() * (j_ * j_.transpose()).inverse();
    nullspace_matrix_ = Eigen::MatrixXd::Identity(n_joints_, n_joints_) - j_inv_ * j_;
    q_dot_cmd_ = j_inv_ * pose_dot_ref_cmd_ + nullspace_matrix_ * q_dot_null_;
    q_cmd_ = q_fbk_ + q_dot_cmd_ * loop_time_;
    for (int i = 0; i < n_joints_; ++i)
    {
        joint_position_msg_.data[i] = q_cmd_(i);
    }
}

void TaskController::update()
{
    if (is_pose_detected_)
    {
        task_pose_publisher_.publish(task_pose_msg_);
    }
    if (!is_initialized_)
    {
        return;
    }
    compute_joint_positions();
    joint_position_publisher_.publish(joint_position_msg_);
    double pos_err = (x_ref_ - x_fbk_).norm();
    double rot_err = tf2::angleShortestPath(quat_fbk_, quat_ref_);
    // ROS_INFO_STREAM("Position error: " << pos_err << " Rotation error: " << rot_err);
}