#include "potential_field_planner/JointPlanner.hpp"

JointPlanner::JointPlanner(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    ROS_INFO_STREAM("Tolerance: " << node_handle_.resolveName("tolerance"));
    if(!load_params())
    {
        ROS_ERROR("Error retrieving parameters");
        ros::requestShutdown();
    }
    joint_velocity_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>("joint_velocities", 1, this);
    joint_position_publisher_ = node_handle_.advertise<std_msgs::Float64MultiArray>("joint_positions", 1, this);
    planning_status_publisher_ = node_handle_.advertise<std_msgs::Bool>("done", 1, this);
    joint_states_subscriber_ = node_handle_.subscribe(joint_states_topic_, 1, &JointPlanner::joint_states_callback, this);
    planner_start_server_ = node_handle_.advertiseService("start", &JointPlanner::planner_start_callback, this);
    // Initialize vectors and arrays
    joint_velocity_cmd_.data.resize(n_joints_);
    joint_position_cmd_.data.resize(n_joints_);
    current_joint_positions_ = Eigen::VectorXd::Zero(n_joints_);
    target_joint_positions_ = current_joint_positions_;
    for (int i = 0; i < n_joints_; ++i)
    {
        target_joint_positions_(i) = joint_target_list_[i];
    }
    is_planner_called_ = false;
    planning_status_msg_.data = false;
    loop_time_ = 1 / loop_rate_;
}

bool JointPlanner::load_params()
{
    if(!node_handle_.getParam("joint_states_topic", joint_states_topic_))
    {
        ROS_ERROR("Could not retrieve parameter joint_states_topic");
        return false;
    }
    if(!node_handle_.getParam("joint_velocity_limits", joint_velocity_limits_))
    {
        ROS_ERROR("Could not retrieve parameter joint_velocity_limits");
        return false;
    }
    if(!node_handle_.getParam("tolerance", tolerance_))
    {
        ROS_ERROR("Could not retrieve parameter tolerance");
        return false;
    }  
    if(!node_handle_.getParam("joint_target_list", joint_target_list_))
    {
        ROS_ERROR("Could not retrieve parameter joint_target_list");
        return false;
    } 
    if(!node_handle_.getParam("k_att", k_att_))
    {
        ROS_ERROR("Could not retrieve parameter k_att");
        return false;
    }
    if(!node_handle_.getParam("n_joints", n_joints_))
    {
        ROS_ERROR("Could not retrieve parameter n_joints");
        return false;
    }
    if(!node_handle_.getParam("/publish_rate", loop_rate_))
    {
        ROS_ERROR("Could not retrieve parameter /publish_rate");
        return false;
    }
    ROS_INFO("Parameters load successfully");
    return true;
}

void JointPlanner::joint_states_callback(const sensor_msgs::JointState& msg)
{
    for (int i = 0; i < n_joints_; ++i)
    {
        current_joint_positions_(i) = msg.position[i];
    }
}

bool JointPlanner::planner_start_callback(std_srvs::TriggerRequest& req, std_srvs::TriggerResponse& resp)
{
    if (is_planner_called_)
    {
        resp.message = "Planner is busy";
        resp.success = false;
        return false;
    }
    is_planner_called_ = true;
    resp.message = "Starting planner";
    resp.success = true;
    return true;
}

void JointPlanner::check_target_reached()
{
    double difference_norm = (target_joint_positions_ - current_joint_positions_).norm();
    if (difference_norm > tolerance_)
    {
        planning_status_msg_.data = false;
    } else
    {
        //is_planner_called_ = false; // Planner target reached, ready to be called again
        planning_status_msg_.data = true;
    }
}

void JointPlanner::compute_field()
{
    Eigen::VectorXd next_joint_velocities = k_att_ * (target_joint_positions_ - current_joint_positions_);
    // ROS_INFO_STREAM("Error: " << (target_joint_positions_ - current_joint_positions_).norm());
    for(int i = 0; i < n_joints_; ++i)
    {
        if(abs(next_joint_velocities(i)) > joint_velocity_limits_[i])
        {
            next_joint_velocities(i) = next_joint_velocities(i) / abs(next_joint_velocities(i)) * joint_velocity_limits_[i];
        }
    }
    Eigen::VectorXd next_joint_positions = current_joint_positions_ + next_joint_velocities * loop_time_;
    for (int i = 0; i < n_joints_; ++i)
    {
        joint_velocity_cmd_.data[i] = next_joint_velocities(i);
        joint_position_cmd_.data[i] = next_joint_positions(i);
    }
}

void JointPlanner::update()
{
    if (is_planner_called_)
    {
        check_target_reached();
        compute_field();
        joint_velocity_publisher_.publish(joint_velocity_cmd_);
        joint_position_publisher_.publish(joint_position_cmd_);
        planning_status_publisher_.publish(planning_status_msg_);
    }
}