#include "potential_field_planner/PotentialFieldPlanner.hpp"

PotentialFieldPlanner::PotentialFieldPlanner(ros::NodeHandle& node_handle) : node_handle_(node_handle)
{
    pose_subscriber_ = node_handle_.subscribe("/husky_velocity_controller/feedback/pose", 1, &PotentialFieldPlanner::pose_callback, this);
    start_planning_server_ = node_handle_.advertiseService("start", &PotentialFieldPlanner::planner_start_callback, this);
    pose_publisher_  = node_handle_.advertise<geometry_msgs::Pose>("pose", 1, this);
    twist_publisher_ = node_handle_.advertise<geometry_msgs::Twist>("twist", 1, this);
    planning_status_publisher_ = node_handle_.advertise<std_msgs::Bool>("done", 1, this);
    ROS_INFO("Started node planner");
    if (!load_params())
    {
        ROS_ERROR("Shutting down thread");
        ros::requestShutdown();
    }
    is_planner_called_ = false;
    loop_time_ = 1 / loop_rate_;
}

bool PotentialFieldPlanner::load_params()
{
    if(!node_handle_.getParam("/target/x", target_position_(0)))
    {
        ROS_ERROR("Could not retrieve parameter /target/x");
        return false;
    }
    if(!node_handle_.getParam("/target/y", target_position_(1)))
    {
        ROS_ERROR("Could not retrieve parameter /target/y");
        return false;
    }
    if(!node_handle_.getParam("/planner/k_att", k_att_))
    {
        ROS_ERROR("Could not retrieve parameter /planner/k_att");
        return false;
    }
    if(!node_handle_.getParam("/husky_velocity_controller/linear/x/max_velocity", max_linear_velocity_))
    {
        ROS_ERROR("Could not retrieve parameter /husky_velocity_controller/linear/x/max_velocity");
        return false;
    }
    if(!node_handle_.getParam("/husky_velocity_controller/angular/z/max_velocity", max_angular_velocity_))
    {
        ROS_ERROR("Could not retrieve parameter /husky_velocity_controller/angular/z/max_velocity");
        return false;
    }
    if (!node_handle_.getParam("/base_controller/loop_rate", loop_rate_))
    {
        ROS_ERROR("Could not retrieve parameter /base_controller/loop_rate");
        return false;
    }
    ROS_INFO("Parameters loaded!");
    return true;
}

void PotentialFieldPlanner::pose_callback(const geometry_msgs::Pose& msg)
{
    current_position_(0) = msg.position.x;
    current_position_(1) = msg.position.y;
}

bool PotentialFieldPlanner::planner_start_callback(std_srvs::TriggerRequest& req, std_srvs::TriggerResponse& resp)
{
    is_planner_called_ = true;
    resp.success = true;
    resp.message = "Planner trigger succeeded.";
    return true;
}

void PotentialFieldPlanner::check_target_reached()
{
    Eigen::Vector2d displacement = target_position_ - current_position_;
    double distance_to_target = displacement.norm();
    if (distance_to_target < 0.1)
    {   
        planning_status_msg_.data = true;
    } else
    {
        planning_status_msg_.data = false;
    }
}

void PotentialFieldPlanner::compute_field()
{
    Eigen::Vector2d displacement = target_position_ - current_position_;
    double field_speed = k_att_ * displacement.norm(); // Velocity towards target
    if (field_speed > max_linear_velocity_)
    {
        field_speed = max_linear_velocity_;
    }
    Eigen::Vector2d field_velocity = displacement / displacement.norm() * field_speed;
    twist_msg_.linear.x = field_velocity(0);
    twist_msg_.linear.y = field_velocity(1);
    // Compute positions
    pose_msg_.position.x = current_position_(0) + field_velocity(0) * loop_time_;
    pose_msg_.position.y = current_position_(1) + field_velocity(1) * loop_time_;
}

void PotentialFieldPlanner::update()
{
    if(is_planner_called_)
    {
        check_target_reached();
        compute_field();
        planning_status_publisher_.publish(planning_status_msg_);
        twist_publisher_.publish(twist_msg_);
        pose_publisher_.publish(pose_msg_);
    }
}